@@ -7,11 +7,13 @@ https://github.com/telegramdesktop/tdesktop/blob/master/LEGAL
 */
 #include "core/application.h"
 
+#include "data/data_abstract_structure.h"
 #include "data/data_photo.h"
 #include "data/data_document.h"
 #include "data/data_session.h"
 #include "data/data_user.h"
 #include "base/timer.h"
+#include "base/event_filter.h"
 #include "base/concurrent_timer.h"
 #include "base/qt_signal_producer.h"
 #include "base/unixtime.h"
@@ -23,8 +25,9 @@ https://github.com/telegramdesktop/tdesktop/blob/master/LEGAL
 #include "core/ui_integration.h"
 #include "chat_helpers/emoji_keywords.h"
 #include "chat_helpers/stickers_emoji_image_loader.h"
-#include "base/platform/base_platform_info.h"
+#include "base/platform/base_platform_url_scheme.h"
 #include "base/platform/base_platform_last_input.h"
+#include "base/platform/base_platform_info.h"
 #include "platform/platform_specific.h"
 #include "mainwindow.h"
 #include "dialogs/dialogs_entry.h"
@@ -32,6 +35,7 @@ https://github.com/telegramdesktop/tdesktop/blob/master/LEGAL
 #include "apiwrap.h"
 #include "api/api_updates.h"
 #include "calls/calls_instance.h"
+#include "countries/countries_manager.h"
 #include "lang/lang_file_parser.h"
 #include "lang/lang_translator.h"
 #include "lang/lang_cloud_manager.h"
@@ -39,10 +43,12 @@ https://github.com/telegramdesktop/tdesktop/blob/master/LEGAL
 #include "lang/lang_instance.h"
 #include "mainwidget.h"
 #include "core/file_utilities.h"
+#include "core/crash_reports.h"
 #include "main/main_account.h"
 #include "main/main_domain.h"
 #include "main/main_session.h"
 #include "media/view/media_view_overlay_widget.h"
+#include "media/view/media_view_open_common.h"
 #include "mtproto/mtproto_dc_options.h"
 #include "mtproto/mtproto_config.h"
 #include "mtproto/mtp_instance.h"
@@ -56,14 +62,17 @@ https://github.com/telegramdesktop/tdesktop/blob/master/LEGAL
 #include "window/window_lock_widgets.h"
 #include "history/history_location_manager.h"
 #include "ui/widgets/tooltip.h"
+#include "ui/gl/gl_detection.h"
 #include "ui/image/image.h"
 #include "ui/text/text_options.h"
 #include "ui/emoji_config.h"
 #include "ui/effects/animations.h"
+#include "ui/cached_round_corners.h"
 #include "storage/serialize_common.h"
 #include "storage/storage_domain.h"
 #include "storage/storage_databases.h"
 #include "storage/localstorage.h"
+#include "payments/payments_checkout_process.h"
 #include "export/export_manager.h"
 #include "window/window_session_controller.h"
 #include "window/window_controller.h"
@@ -73,7 +82,6 @@ https://github.com/telegramdesktop/tdesktop/blob/master/LEGAL
 #include "boxes/confirm_phone_box.h"
 #include "boxes/confirm_box.h"
 #include "boxes/share_box.h"
-#include "facades.h"
 #include "app.h"
 
 #include <QtWidgets/QDesktopWidget>
@@ -88,6 +96,27 @@ constexpr auto kQuitPreventTimeoutMs = crl::time(1500);
 constexpr auto kAutoLockTimeoutLateMs = crl::time(3000);
 constexpr auto kClearEmojiImageSourceTimeout = 10 * crl::time(1000);
 
+void SetCrashAnnotationsGL() {
+#ifdef Q_OS_WIN
+	CrashReports::SetAnnotation("OpenGL ANGLE", [] {
+		if (Core::App().settings().disableOpenGL()) {
+			return "Disabled";
+		} else switch (Ui::GL::CurrentANGLE()) {
+		case Ui::GL::ANGLE::Auto: return "Auto";
+		case Ui::GL::ANGLE::D3D11: return "Direct3D 11";
+		case Ui::GL::ANGLE::D3D9: return "Direct3D 9";
+		case Ui::GL::ANGLE::D3D11on12: return "D3D11on12";
+		case Ui::GL::ANGLE::OpenGL: return "OpenGL";
+		}
+		Unexpected("Ui::GL::CurrentANGLE value in SetupANGLE.");
+	}());
+#else // Q_OS_WIN
+	CrashReports::SetAnnotation(
+		"OpenGL",
+		Core::App().settings().disableOpenGL() ? "Disabled" : "Enabled");
+#endif // Q_OS_WIN
+}
+
 } // namespace
 
 Application *Application::Instance = nullptr;
@@ -150,6 +179,15 @@ Application::~Application() {
 	_window = nullptr;
 	_mediaView = nullptr;
 	_notifications->clearAllFast();
+
+	// We must manually destroy all windows before going further.
+	// DestroyWindow on Windows (at least with an active WebView) enters
+	// event loop and invoke scheduled crl::on_main callbacks.
+	//
+	// For example Domain::removeRedundantAccounts() is called from
+	// Domain::finish() and there is a violation on Ensures(started()).
+	Payments::CheckoutProcess::ClearAll();
+
 	_domain->finish();
 
 	Local::finish();
@@ -159,14 +197,14 @@ Application::~Application() {
 	Ui::Emoji::Clear();
 	Media::Clip::Finish();
 
-	App::deinitMedia();
+	Ui::FinishCachedCorners();
+	Data::clearGlobalStructures();
 
 	Window::Theme::Uninitialize();
 
 	Media::Player::finish(_audio.get());
 	style::stopManager();
 
-	Global::finish();
 	ThirdParty::finish();
 
 	Instance = nullptr;
@@ -176,8 +214,7 @@ void Application::run() {
 	style::internal::StartFonts();
 
 	ThirdParty::start();
-	Global::start();
-	refreshGlobalProxy(); // Depends on Global::start().
+	refreshGlobalProxy(); // Depends on Core::IsAppLaunched().
 
 	// Depends on OpenSSL on macOS, so on ThirdParty::start().
 	// Depends on notifications settings.
@@ -187,6 +224,9 @@ void Application::run() {
 	ValidateScale();
 
 	if (Local::oldSettingsVersion() < AppVersion) {
+#ifndef DESKTOP_APP_USE_PACKAGED
+		RegisterUrlScheme();
+#endif // DESKTOP_APP_USE_PACKAGED
 		psNewVersion();
 	}
 
@@ -206,6 +246,7 @@ void Application::run() {
 
 	style::startManager(cScale());
 	Ui::InitTextOptions();
+	Ui::StartCachedCorners();
 	Ui::Emoji::Init();
 	startEmojiImageLoader();
 	startSystemDarkModeViewer();
@@ -251,14 +292,13 @@ void Application::run() {
 
 	// Depend on activeWindow() for now :(
 	startShortcuts();
-	App::initMedia();
 	startDomain();
 
 	_window->widget()->show();
 
 	const auto currentGeometry = _window->widget()->geometry();
 	_mediaView = std::make_unique<Media::View::OverlayWidget>();
-	_window->widget()->setGeometry(currentGeometry);
+	_window->widget()->Ui::RpWidget::setGeometry(currentGeometry);
 
 	DEBUG_LOG(("Application Info: showing."));
 	_window->finishFirstShow();
@@ -272,6 +312,51 @@ void Application::run() {
 	for (const auto &error : Shortcuts::Errors()) {
 		LOG(("Shortcuts Error: %1").arg(error));
 	}
+
+	SetCrashAnnotationsGL();
+	if (!Platform::IsMac() && Ui::GL::LastCrashCheckFailed()) {
+		showOpenGLCrashNotification();
+	}
+
+	_window->openInMediaViewRequests(
+	) | rpl::start_with_next([=](Media::View::OpenRequest &&request) {
+		if (_mediaView) {
+			_mediaView->show(std::move(request));
+		}
+	}, _window->lifetime());
+
+	{
+		const auto countries = std::make_shared<Countries::Manager>(
+			_domain.get());
+		countries->lifetime().add([=] {
+			[[maybe_unused]] const auto countriesCopy = countries;
+		});
+	}
+}
+
+void Application::showOpenGLCrashNotification() {
+	const auto enable = [=] {
+		Ui::GL::ForceDisable(false);
+		Ui::GL::CrashCheckFinish();
+		Core::App().settings().setDisableOpenGL(false);
+		Local::writeSettings();
+		App::restart();
+	};
+	const auto keepDisabled = [=] {
+		Ui::GL::ForceDisable(true);
+		Ui::GL::CrashCheckFinish();
+		Core::App().settings().setDisableOpenGL(true);
+		Local::writeSettings();
+	};
+	_window->show(Box<ConfirmBox>(
+		"There may be a problem with your graphics drivers and OpenGL. "
+		"Try updating your drivers.\n\n"
+		"OpenGL has been disabled. You can try to enable it again "
+		"or keep it disabled if crashes continue.",
+		"Enable",
+		"Keep Disabled",
+		enable,
+		keepDisabled));
 }
 
 void Application::startDomain() {
@@ -281,8 +366,6 @@ void Application::startDomain() {
 		startSettingsAndBackground();
 	}
 	if (state != Storage::StartResult::Success) {
-		Global::SetLocalPasscode(true);
-		Global::RefLocalPasscodeChanged().notify();
 		lockByPasscode();
 		DEBUG_LOG(("Application Info: passcode needed..."));
 	}
@@ -347,66 +430,6 @@ bool Application::hideMediaView() {
 	return false;
 }
 
-void Application::showPhoto(not_null<const PhotoOpenClickHandler*> link) {
-	const auto photo = link->photo();
-	const auto peer = link->peer();
-	const auto item = photo->owner().message(link->context());
-	return (!item && peer)
-		? showPhoto(photo, peer)
-		: showPhoto(photo, item);
-}
-
-void Application::showPhoto(not_null<PhotoData*> photo, HistoryItem *item) {
-	Expects(_mediaView != nullptr);
-
-	_mediaView->showPhoto(photo, item);
-	_mediaView->activateWindow();
-	_mediaView->setFocus();
-}
-
-void Application::showPhoto(
-		not_null<PhotoData*> photo,
-		not_null<PeerData*> peer) {
-	Expects(_mediaView != nullptr);
-
-	_mediaView->showPhoto(photo, peer);
-	_mediaView->activateWindow();
-	_mediaView->setFocus();
-}
-
-void Application::showDocument(not_null<DocumentData*> document, HistoryItem *item) {
-	Expects(_mediaView != nullptr);
-
-	if (cUseExternalVideoPlayer()
-		&& document->isVideoFile()
-		&& !document->filepath().isEmpty()) {
-		File::Launch(document->location(false).fname);
-	} else {
-		_mediaView->showDocument(document, item);
-		_mediaView->activateWindow();
-		_mediaView->setFocus();
-	}
-}
-
-void Application::showTheme(
-		not_null<DocumentData*> document,
-		const Data::CloudTheme &cloud) {
-	Expects(_mediaView != nullptr);
-
-	_mediaView->showTheme(document, cloud);
-	_mediaView->activateWindow();
-	_mediaView->setFocus();
-}
-
-PeerData *Application::ui_getPeerForMouseAction() {
-	if (_mediaView && !_mediaView->isHidden()) {
-		return _mediaView->ui_getPeerForMouseAction();
-	} else if (const auto m = App::main()) { // multi good
-		return m->ui_getPeerForMouseAction();
-	}
-	return nullptr;
-}
-
 bool Application::eventFilter(QObject *object, QEvent *e) {
 	switch (e->type()) {
 	case QEvent::KeyPress:
@@ -493,17 +516,17 @@ void Application::setCurrentProxy(
 		const MTP::ProxyData &proxy,
 		MTP::ProxyData::Settings settings) {
 	const auto current = [&] {
-		return (Global::ProxySettings() == MTP::ProxyData::Settings::Enabled)
-			? Global::SelectedProxy()
+		return _settings.proxy().isEnabled()
+			? _settings.proxy().selected()
 			: MTP::ProxyData();
 	};
 	const auto was = current();
-	Global::SetSelectedProxy(proxy);
-	Global::SetProxySettings(settings);
+	_settings.proxy().setSelected(proxy);
+	_settings.proxy().setSettings(settings);
 	const auto now = current();
 	refreshGlobalProxy();
 	_proxyChanges.fire({ was, now });
-	Global::RefConnectionTypeChanged().notify();
+	_settings.proxy().connectionTypeChangesNotify();
 }
 
 auto Application::proxyChanges() const -> rpl::producer<ProxyChange> {
@@ -511,11 +534,10 @@ auto Application::proxyChanges() const -> rpl::producer<ProxyChange> {
 }
 
 void Application::badMtprotoConfigurationError() {
-	if (Global::ProxySettings() == MTP::ProxyData::Settings::Enabled
-		&& !_badProxyDisableBox) {
+	if (_settings.proxy().isEnabled() && !_badProxyDisableBox) {
 		const auto disableCallback = [=] {
 			setCurrentProxy(
-				Global::SelectedProxy(),
+				_settings.proxy().selected(),
 				MTP::ProxyData::Settings::System);
 		};
 		_badProxyDisableBox = Ui::show(Box<InformBox>(
@@ -527,6 +549,9 @@ void Application::badMtprotoConfigurationError() {
 void Application::startLocalStorage() {
 	Local::start();
 	_saveSettingsTimer.emplace([=] { saveSettings(); });
+	_settings.saveDelayedRequests() | rpl::start_with_next([=] {
+		saveSettingsDelayed();
+	}, _lifetime);
 }
 
 void Application::startEmojiImageLoader() {
@@ -557,6 +582,14 @@ void Application::startEmojiImageLoader() {
 	}, _lifetime);
 }
 
+void Application::setScreenIsLocked(bool locked) {
+	_screenIsLocked = locked;
+}
+
+bool Application::screenIsLocked() const {
+	return _screenIsLocked;
+}
+
 void Application::setDefaultFloatPlayerDelegate(
 		not_null<Media::Player::FloatDelegate*> delegate) {
 	Expects(!_defaultFloatPlayerDelegate == !_floatPlayers);
@@ -847,7 +880,7 @@ bool Application::passcodeLocked() const {
 void Application::updateNonIdle() {
 	_lastNonIdleTime = crl::now();
 	if (const auto session = maybeActiveSession()) {
-		session->updates().checkIdleFinish();
+		session->updates().checkIdleFinish(_lastNonIdleTime);
 	}
 }
 
@@ -874,19 +907,21 @@ bool Application::someSessionExists() const {
 	return false;
 }
 
-void Application::checkAutoLock() {
-	if (!Global::LocalPasscode()
+void Application::checkAutoLock(crl::time lastNonIdleTime) {
+	if (!_domain->local().hasLocalPasscode()
 		|| passcodeLocked()
 		|| !someSessionExists()) {
 		_shouldLockAt = 0;
 		_autoLockTimer.cancel();
 		return;
+	} else if (!lastNonIdleTime) {
+		lastNonIdleTime = this->lastNonIdleTime();
 	}
 
 	checkLocalTime();
 	const auto now = crl::now();
 	const auto shouldLockInMs = _settings.autoLock() * 1000LL;
-	const auto checkTimeMs = now - lastNonIdleTime();
+	const auto checkTimeMs = now - lastNonIdleTime;
 	if (checkTimeMs >= shouldLockInMs || (_shouldLockAt > 0 && now > _shouldLockAt + kAutoLockTimeoutLateMs)) {
 		_shouldLockAt = 0;
 		_autoLockTimer.cancel();
@@ -908,7 +943,7 @@ void Application::checkAutoLockIn(crl::time time) {
 void Application::localPasscodeChanged() {
 	_shouldLockAt = 0;
 	_autoLockTimer.cancel();
-	checkAutoLock();
+	checkAutoLock(crl::now());
 }
 
 bool Application::hasActiveWindow(not_null<Main::Session*> session) const {
@@ -965,10 +1000,10 @@ bool Application::minimizeActiveWindow() {
 }
 
 QWidget *Application::getFileDialogParent() {
-	return (_mediaView && _mediaView->isVisible())
-		? (QWidget*)_mediaView.get()
+	return (_mediaView && !_mediaView->isHidden())
+		? static_cast<QWidget*>(_mediaView->widget())
 		: activeWindow()
-		? (QWidget*)activeWindow()->widget()
+		? static_cast<QWidget*>(activeWindow()->widget())
 		: nullptr;
 }
 
@@ -978,18 +1013,9 @@ void Application::notifyFileDialogShown(bool shown) {
 	}
 }
 
-QWidget *Application::getModalParent() {
-	return (Platform::IsWayland() && activeWindow())
-		? activeWindow()->widget().get()
-		: nullptr;
-}
-
-
 void Application::checkMediaViewActivation() {
 	if (_mediaView && !_mediaView->isHidden()) {
-		_mediaView->activateWindow();
-		QApplication::setActiveWindow(_mediaView.get());
-		_mediaView->setFocus();
+		_mediaView->activate();
 	}
 }
 
@@ -1003,30 +1029,49 @@ QPoint Application::getPointForCallPanelCenter() const {
 // macOS Qt bug workaround, sometimes no leaveEvent() gets to the nested widgets.
 void Application::registerLeaveSubscription(not_null<QWidget*> widget) {
 #ifdef Q_OS_MAC
-	if (const auto topLevel = widget->window()) {
-		if (topLevel == _window->widget()) {
-			auto weak = Ui::MakeWeak(widget);
-			auto subscription = _window->widget()->leaveEvents(
-			) | rpl::start_with_next([weak] {
-				if (const auto window = weak.data()) {
-					QEvent ev(QEvent::Leave);
-					QGuiApplication::sendEvent(window, &ev);
+	if (const auto window = widget->window()) {
+		auto i = _leaveFilters.find(window);
+		if (i == end(_leaveFilters)) {
+			const auto check = [=](not_null<QEvent*> e) {
+				if (e->type() == QEvent::Leave) {
+					if (const auto taken = _leaveFilters.take(window)) {
+						for (const auto &weak : taken->registered) {
+							if (const auto widget = weak.data()) {
+								QEvent ev(QEvent::Leave);
+								QCoreApplication::sendEvent(widget, &ev);
+							}
+						}
+						delete taken->filter.data();
+					}
 				}
+				return base::EventFilterResult::Continue;
+			};
+			const auto filter = base::install_event_filter(window, check);
+			QObject::connect(filter, &QObject::destroyed, [=] {
+				_leaveFilters.remove(window);
 			});
-			_leaveSubscriptions.emplace_back(weak, std::move(subscription));
+			i = _leaveFilters.emplace(
+				window,
+				LeaveFilter{ .filter = filter.get() }).first;
 		}
+		i->second.registered.push_back(widget.get());
 	}
 #endif // Q_OS_MAC
 }
 
 void Application::unregisterLeaveSubscription(not_null<QWidget*> widget) {
 #ifdef Q_OS_MAC
-	_leaveSubscriptions = std::move(
-		_leaveSubscriptions
-	) | ranges::action::remove_if([&](const LeaveSubscription &subscription) {
-		auto pointer = subscription.pointer.data();
-		return !pointer || (pointer == widget);
-	});
+	if (const auto topLevel = widget->window()) {
+		const auto i = _leaveFilters.find(topLevel);
+		if (i != end(_leaveFilters)) {
+			i->second.registered = std::move(
+				i->second.registered
+			) | ranges::actions::remove_if([&](QPointer<QWidget> widget) {
+				const auto pointer = widget.data();
+				return !pointer || (pointer == widget);
+			});
+		}
+	}
 #endif // Q_OS_MAC
 }
 
@@ -1101,7 +1146,7 @@ void Application::startShortcuts() {
 			return true;
 		});
 		request->check(Command::Lock) && request->handle([=] {
-			if (!passcodeLocked() && Global::LocalPasscode()) {
+			if (!passcodeLocked() && _domain->local().hasLocalPasscode()) {
 				lockByPasscode();
 				return true;
 			}
@@ -1116,6 +1161,19 @@ void Application::startShortcuts() {
 	}, _lifetime);
 }
 
+void Application::RegisterUrlScheme() {
+	base::Platform::RegisterUrlScheme(base::Platform::UrlSchemeDescriptor{
+		.executable = cExeDir() + cExeName(),
+		.arguments = qsl("-workdir \"%1\"").arg(cWorkingDir()),
+		.protocol = qsl("tg"),
+		.protocolName = qsl("Telegram Link"),
+		.shortAppName = qsl("tdesktop"),
+		.longAppName = QCoreApplication::applicationName(),
+		.displayAppName = AppName.utf16(),
+		.displayAppDescription = AppName.utf16(),
+	});
+}
+
 bool IsAppLaunched() {
 	return (Application::Instance != nullptr);
 }
